/* ########################################################################

			     Smac file ada mode

   File: ada.sc
   Path: /mnt/LaCie/xcoral-3.49.61/SmacLib/ada.sc
   Description: 
   Created: sam. mai  2 15:51:30 CEST 2020
   Author: lionel fournigault
   Modified: sam. mai  2 15:51:30 CEST 2020
   Last maintained by: lionel fournigault

   RCS $Revision$ $State$
   

   ########################################################################

   Note: ada_color_region is defined in Smac file color.sc

   ada_new_line: return and indent
   ada_indent_line: Tab key to indent line
   ada_indent_region: use mark to indent rgion
	 
########################################################################

   Copyright (c) : lionel fournigault

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

   ######################################################################## */

int ada_indent_step = 4;

char * ada_regexp_indent = "select\\|loop\\|begin\\|when\\|task\\|procedure\
\\|private\\|while\\|for\\|do\\|or\\|protected\\|elseif\\|if\\|package\\|case";
char * ada_regexp_indent2 = "or\\|begin\\|else\\|elseif";

char * ada_string = "\'.*\'\\|\".*\""; /* replace gen_string */

/* ================================================================================
   Return line number at position. First line of buffer = 0
   ================================================================================ */
int ada_no_line_at_pos ( int position ) {

  int pos = 0, n = 0;
		     
  while(pos != end_of_file()) {
    if (pos == position){
      return n;
    }
    if (the_char(pos) == '\n')
      n += 1;
    pos += 1;
  }
  return n;
}

/* ================================================================================
   Search position inside regexp. Return 1 if position is inside regexp else 0
   ================================================================================ */
int ada_pos_in_regexp ( char *regexp, int nsub, int position ) {

  int orig = current_position();
  int pos, re_begin, re_end;

  goto_char(0);
  while (1) {
    pos = re_forward_search (regexp);
    if ((pos == -1) || (pos == -2) || (pos >= end_of_file()))
      break;
    
    if (pos > position){
      goto_char (orig);
      return 0;
    }
    
    re_begin = re_match_beginning(nsub);
    re_end = re_match_end(nsub);
    
    if ( position >= re_begin && position <= re_end ) {
      /* Bingo */
      goto_char (orig);
      return 1;
    }

    if (re_end == end_of_file())
      break;
    else
      goto_char(re_end + 1);
  }
  goto_char (orig);
  return 0;
}

/* ================================================================================
   Search blinkchar backward in the buffer. Return position or -1 
   ================================================================================ */
int ada_pos_blink_char ( int blink_char, int c ) {
  
  int pos, c_char, pos_blink = -1, internal = 0;
  
  pos = current_position();
  
  if (pos < 2)
    return;
  else
    pos -= 2;
  
  if (pos == 0 && the_char(pos) == blink_char)
    pos_blink = 0;
  
  while ( pos >= 0 ){
    c_char = the_char(pos);
    
    if ( c_char == blink_char ) {
      if ( ada_pos_in_regexp ( re_ada_comment, 0, pos ) == 0 
	  && ada_pos_in_regexp ( ada_string, 0, pos ) == 0 ) {
	if (internal == 0) {
	  pos_blink = pos;
	  break;
	}
	else {
	  internal -= 1;
	}
      }
    }
    else if (c_char == c ) {
      if ( ada_pos_in_regexp ( re_ada_comment, 0, pos ) == 0 
	  && ada_pos_in_regexp ( ada_string, 0, pos ) == 0 ) {
	internal += 1;
      }
    }
    else
      ;
    pos -= 1;
  }
  return pos_blink;
}

/* ================================================================================
   Align two lines. Use after blink;
   ================================================================================ */
void ada_align_lines ( int ref_pos, int orig ) {

  int pos, c;
  int count = 0;

  if (ada_no_line_at_pos(ref_pos) == ada_no_line_at_pos(orig))
    return;

  /* line to align */
  goto_beginning_of_line();
  pos = current_position();
  while (pos != end_of_line()){
    if ( the_char(pos) == ' ' || the_char(pos) == '\t' ) {
      delete_char();
      count -= 1;
      pos = current_position();
    }
    else
      break;
  }

  /* Reference line */
  pos = ref_pos;
  if (the_char(pos) == '\n')
    pos -= 1;
  
  while (1) {
    if (the_char(pos) == '\n'){
      pos += 1;
      break;
    }
    if (pos == 0)
      break;
    pos -= 1;
  }
  
  while (1) {
    c = the_char(pos);
    if ( c == ' ' || c == '\t' ) {
      insert_char(c);
      count += 1;
    }
    else
      break;
    pos += 1;
  }
  goto_char ( orig + count );
}

/* ================================================================================
   Return first character in line (not blanck space or tab)
   ================================================================================ */
char ada_first_char_in_line ( int line ) {
  
  int pos = current_position();
  int tmp_pos;
  
  if (current_line() != line)
    goto_line(line);
  
  if ( beginning_of_line() ==  end_of_line() ){
    goto_char (pos);
    return '\n';
  }
  else
    goto_beginning_of_line();

  while (1){
    if (current_position() == end_of_line()){
     goto_char (pos);
     return '\n';
    }
    if (the_char(current_position()) != ' ' 
	&& the_char(current_position()) != '\t'){
      tmp_pos = current_position();
      goto_char (pos);
      return the_char(tmp_pos);
    }
    else
      goto_next_char();
  }
}

/* ================================================================================
   Return last character in current line (not blank space or tab)
   Ignore single line comment
   ================================================================================ */
char ada_last_char_in_line() {
  
  int pos = end_of_line();
  
  if (pos == beginning_of_line()) /* empty line */
      return '\n';
  else
    pos -= 1;
      
  while (1) {
    if ( pos == beginning_of_line() )
      break;
    if (the_char(pos) == ' ' || the_char(pos)== '\t'
	|| ada_pos_in_regexp ( re_ada_comment, 0, pos ) == 1 )
      pos -= 1;
    else
      break;
  }
  return the_char(pos);
}

/* ================================================================================
   Return distance from end of line. 
   ================================================================================ */
int ada_distance_from_end_of_line (int pos) {
  
  int distance = 0, tmp_pos;
  
  if (end_of_line() == beginning_of_line())
    distance = 0;
    
  tmp_pos = end_of_line();
  while (tmp_pos != beginning_of_line()){
    if (pos == tmp_pos)
      break;
    else
      distance += 1;
    tmp_pos -= 1;
  }
  /* a partir d'ici tmp_pos = pos ou debut de ligne */
  tmp_pos = beginning_of_line();
  while (tmp_pos != end_of_line()){
    if (the_char(tmp_pos) != ' ' && the_char(tmp_pos) != '\t'){
      break;
    }
    else {
      tmp_pos += 1;
     }
  }
  if (tmp_pos > pos)
    distance -= tmp_pos - pos;

  return distance;
}

/* ================================================================================
   Align to previous line.
   ================================================================================ */
void ada_align_to_previous_line() {
  
  int pos, c;

  if (current_line() == 0)
    return;
  /* Remove ' ' or tab at the beginning of line */
  goto_beginning_of_line();
  while(current_char()==' ' || current_char()=='\t'){
    if (current_position() == end_of_line())
      break;
    delete_char();
  }
  
  /* Use indent of previous line */  
  goto_previous_line();
  goto_beginning_of_line();
  pos = current_position();
  goto_next_line();
  goto_beginning_of_line();
  
  while (strchr(" \t", c = the_char(pos))) {
    if (pos == end_of_line())
      break;
    insert_char(c);
    pos += 1;
  } 
}

/* ================================================================================
   Return pos blink line. Multi-lines parantheses before blink
   ================================================================================ */
int ada_pos_line_blink (int orig) {

  int tmp_pos,  pos = orig;
  int pos_blink = pos;

  goto_char(orig);
  /* on est sur le blink caractre { */
  /* recherche d'une parenthse fermante dans la ligne courante */

  while (1){
    goto_previous_char();
    tmp_pos = current_position();
    if (tmp_pos == beginning_of_line())
      break;
    if ( the_char(tmp_pos) == ' ' || the_char(tmp_pos) == '\t' )
      continue;
    else
      break;
  }

  if (the_char(tmp_pos) == ')') {
    goto_char(tmp_pos+1);   
    pos_blink = ada_pos_blink_char ('(', ')');
    if (pos_blink == -1)
      pos_blink = pos;
  }
  goto_char(orig);
  return pos_blink;
}

/* ================================================================================
   Indent more if need
   ================================================================================ */
int ada_indent_more_if_need () {
  
  int first_char_position, n;
  int more_indent = 0;
  
  /* Move to previous line at the first character not blank space or tab */
  goto_previous_line();
  goto_beginning_of_line();
  
  while(current_char()==' ' || current_char()=='\t'){
    if (current_position() == end_of_line())
      break;
    goto_next_char();
  }
  
  first_char_position = current_position();
  /* indent more if need */
  if (((re_forward_search(ada_regexp_indent) == first_char_position)
       && ada_pos_in_regexp (re_ada_comment, 0, first_char_position) == 0 )) {
    more_indent = 1;
    n = ada_indent_step;
    goto_next_line();
    goto_beginning_of_line();
    while (n--){
      insert_char(' ');
    }
  }
  else
    goto_next_line(); /* return to current line */
  return more_indent;
}

/* ================================================================================
   Return no colum of first char not blanck or tab.
   ================================================================================ */
int ada_first_col (int line) {
  
  int count = 0;
  int pos = current_position();
  
  goto_line(line);
  goto_beginning_of_line();
  while ( current_position() != end_of_line()) {
    if (current_char() == ' ' || current_char() == '\t'){
      goto_next_char();
      count += 1;
    }
    else
      break;
  }
  goto_char(pos);
  return count;
}

/* ================================================================================
   Align end keyword
   ================================================================================ */
void ada_align_end_keyword () {
  
  int orig = current_position();
  int pos = orig;
  int len, c, n;
  char * word, *p;

  pos += 3;
  while ( pos != end_of_line() ){
    if ( the_char(pos) == ' ' || the_char(pos) == '\t')
      pos += 1;
    else
      break;
  }
  len = end_of_line() - pos;

  if ( len > 1) {
    word = (char *)malloc(len+1);
    p = word;
    while ( strchr (chars_of_word, c=the_char(pos))){
      if (pos == end_of_line())
	break;
      *p = c;
      pos += 1;
      p += 1;
    }
    *p = '\0';

    while (1){
      pos = re_backward_search(word);					  
      if (pos == -1)
	break;
      if (ada_pos_in_regexp (re_ada_comment, 0, pos) == 0)
	break;
      else
	goto_char(pos);
    }
    if ( pos != -1 ){
      goto_char(orig);
      ada_align_lines(pos, orig);
    }
    else {
      goto_char(orig);
      n = ada_indent_step;
      goto_beginning_of_line();
      while (n--){
	if (current_char()==' ' || current_char()=='\t')
	  delete_char();
	else
	  break;
      }
    }
    free(word);
  }
  else {
    n = ada_indent_step;
    goto_beginning_of_line();
    while (n--){
      if (current_char()==' ' || current_char()=='\t')
	delete_char();
      else
	break;
    }
  }
}

/* ================================================================================
   Indent less if need.
   ================================================================================ */
void ada_indent_less_if_need () {
  
  int first_char_position, pos, pos_org, col_org, col, n;

  pos_org = current_position();
  col_org = ada_first_col(current_line());
  
  goto_beginning_of_line();
  while(current_char()==' ' || current_char()=='\t'){
    if (current_position() == end_of_line())
      break;
    goto_next_char();
  }
  
  first_char_position = current_position();
  
  if ((re_forward_search("end") == first_char_position)){
    ada_align_end_keyword();
    return;
  }
  
  if ((re_forward_search("task ") == first_char_position)){
    /* recherche block definition */
    pos = re_backward_search("procedure ");
    if (pos != -1){
      ada_align_lines ( pos, pos_org );
      goto_char(pos_org);
      goto_beginning_of_line();
      n = ada_indent_step;
      while (n--){
	  insert_char(' ');
      }
    }
  }
  else if ((re_forward_search(ada_regexp_indent2) == first_char_position)){
    goto_beginning_of_line();
    n = ada_indent_step;
    while (n--){
      if (current_char()==' ' || current_char()=='\t')
	delete_char();
      else
	break;
    }
  }
}

/* ================================================================================
   Return last word in line.
   ================================================================================ */
char *ada_last_word_in_line (int line) {
  
  int orig, pos, word = 0;
  char * result, * p;
  
  orig = current_position();
  
  if (current_line() != line)
    goto_line(line);
  
  goto_end_of_line();
  pos = current_position();
  while (pos != beginning_of_line()) {
    if (the_char(pos) == ' ')
      break;
    pos -= 1;
    word += 1;
  }
  pos += 1;
  
  if (word > 0) {
    result = (char *)malloc (word + 1);
    p = result;
    while (pos != end_of_line()) {
      *p = the_char(pos);
      p += 1;
      pos += 1;
    }
    goto_char (orig);
    return result;
  }
  else {
    goto_char (orig);
    return 0;
  }
}

/* ================================================================================
   Indent line. Entry point of main indent process.
   ================================================================================ */
void ada_indent_line ()
{
  int orig = current_position();
  int pos = orig, n, indent_m = 0, c_line, first_char;
  int c, first_char_position, distance = 0, tmp_pos, pos_blink;

  if (current_line()==0)
    return;
  
  distance = ada_distance_from_end_of_line(pos);
  
  /* first align to previous line then indent more if need */
  ada_align_to_previous_line();
  indent_m = ada_indent_more_if_need ();
  
  c_line = current_line();
  /* Indent less if need */
  if (indent_m == 0)
    ada_indent_less_if_need ();
  goto_line (c_line);

  /* Return to initial position */
  if (distance == 0)
    goto_end_of_line();
  else {
    goto_end_of_line();
    while(distance > 0){
      if (current_position() == beginning_of_line())
	break;
      goto_previous_char();
      distance -= 1;
    }
  }
  if (current_buffer_is_colored())
    ada_color_region (beginning_of_line(), end_of_line());
}

/* ================================================================================
   Return and indent.
   ================================================================================ */
void ada_new_line ()
{
  insert_char('\n');
  ada_indent_line();
}

/* ================================================================================
   Indent region
   ================================================================================ */
void ada_indent_region() {

  int init_line = current_line();
  int mark_line, i, start_line, end_line;

  if (mark_position() == -1)
    return;
  
  watch_on();
  goto_mark();
  mark_line = current_line();
  if (mark_line > init_line){
    start_line = init_line;
    end_line = mark_line;
  }
  else {
    start_line = mark_line;
    end_line = init_line;
  }
  goto_line(start_line);

  for (i=start_line; i < end_line; i++ ){
    ada_indent_line();
    goto_next_line();
  }
  ada_indent_line();
  goto_line (init_line);
  goto_beginning_of_line();

  watch_off();
  reset_mark();
}

/* ================================================================================
   Indent buffer
   ================================================================================ */
void ada_indent_buffer() {
  
  int line = current_line();
  
  set_mark(0);
  goto_end_of_file();
  
  ada_indent_region();
  
  goto_line(line);
  goto_beginning_of_line();
  color_buffer();
  
  reset_mark();
}

/* ================================================================================
   Insert the char " ' and colorize
   ================================================================================ */
void ada_insert_and_colorize ()
{
  insert_char(last_key());
  
  if (current_buffer_is_colored())
    ada_color_region (beginning_of_line(), end_of_line());
}

/* ================================================================================
   Insert, blink matched char and indent if need
   ================================================================================ */
void ada_insert_and_blink() {
  
  int orig, pos, tmp_pos;
  int c, c_char, blink_char, internal = 0;
  
  c = last_key();

  if (c == ']')
    blink_char = '[';
  else if (c == ')')
    blink_char = '(';
  else
    return;
   
  insert_char (c);
  orig = current_position();
  pos = orig;
  
  /* Check inside comment or string */ 
  if ( ada_pos_in_regexp ( re_ada_comment, 0, pos ) == 1 
      || ada_pos_in_regexp ( ada_string, 0, pos ) == 1 ) {
    return;
  }

  pos = ada_pos_blink_char (blink_char, c);

  if ( pos == -1)
    return;
  else
    blink(pos);
}

/* ================================================================================
   Ada key_def
   ================================================================================ */
{
  key_def("Ada", "\t", "ada_indent_line");
  key_def("Ada", "\r", "ada_new_line");
  key_def("Ada", " ", "ada_insert_and_colorize");
  
  key_def("Ada", "\"", "ada_insert_and_colorize");
  key_def("Ada", "\'", "ada_insert_and_colorize");

  key_def("Ada", "]", "ada_insert_and_blink");
  key_def("Ada", ")", "ada_insert_and_blink");

  key_def("Ada", "^[i", "ada_indent_region");
  
  key_def("Ada", "^[u", "upcase_word");
  key_def("Ada", "^[l", "downcase_word");
  key_def("Ada", "^[c", "capitalize_word");
}
